import os
import platform
import re
import sys
import time

from bloonix.socket.serialized import Client
from bloonix.dispatcher import Dispatcher
from bloonix.logger import logger, Logger
from bloonix.plugin import execute_check
from bloonix import procutils
import bloonix.utils as utils

class Agent:
    VERSION = "1.21"

    def __init__(
            self, config_file, pid_file,
            background=True, log_level="debug"):

        self.config_file = config_file
        self.pid_file = pid_file
        self.config = Validate.main_config(config_file)

        if not os.path.isdir(self.config["plugin_cache_dir"]):
            os.makedirs(self.config["plugin_cache_dir"], mode=0o755)
            if self.config["user"] is not None:
                os.chown(
                    self.config["plugin_cache_dir"],
                    procutils.get_uid(self.config["user"]),
                    procutils.get_gid(self.config["group"])
                )

        if background is True:
            procutils.hang_up(
                self.config["user"],
                self.config["group"],
                self.pid_file)
        else:
            procutils.switch_to(
                self.config["user"],
                self.config["group"])
            self.config["logger"] = {
                "screen": {
                    "level": log_level,
                    "message_layout": "%T %L %P %t %m",
                    "timeformat": "%H:%M:%S"
                }
            }

        for k, v in self.config["environ"].items():
            if (isinstance(k, str)
                    and isinstance(v, str)
                    and re.search(r"^[A-Z0-9_]+", k)):
                os.environ[k] = v

        self.log = logger("bloonix")
        self.log.config(self.config["logger"])
        self.hosts_by_id = {}
        self.log.notice("bloonix-agent version %s started" % self.VERSION)
        self.log.notice(
            "available compression methods:",
            Client.get_available_compression_methods())

        # Hosts that are ready for the next check are queued first.
        # New hosts are appended to the end with append().
        # The host to be processed next is pulled out from the
        # front with pop(0).
        self.ready_hosts = []

        if "host" in self.config and self.config["host"] is not None:
            self.save_host(self.config["host"], False)

        if self.config["location_group"] is not None:
            self.config["location_group"]["last_check"] = 0

        if self.config["testmode"]:
            self.test_data = test_data()
        else:
            self.test_data = None

        self.check_uptime()
        self.dispatcher = Dispatcher(self.config["worker"])
        self.dispatcher.on("init", self.on_init)
        self.dispatcher.on("ready", self.on_ready)
        self.dispatcher.on("process", self.on_process)
        self.dispatcher.on("finish", self.on_finish)

    def check_uptime(self):
        if platform.system() != "Linux":
            return
        try:
            f = open("/proc/uptime", "r")
            m = re.search(r"^(\d+)", f.readline())
            uptime = int(m.group(1))
            f.close()
            if uptime < 30:
                s = 30 - uptime
                self.log.notice(
                    "the system was recently booted, sleep for %s seconds"
                    % s)
                time.sleep(s)
        except Exception: pass

    def startup(self):
        self.dispatcher.run()

    def save_host(self, host, location_group):
        host_id = host["host_id"]
        self.log.info("save new host with id %s" % host_id)
        self.hosts_by_id[host_id] = Host(host, location_group)

    # Executed in parent.
    # Check if hosts are ready.
    def on_ready(self, ready_worker):
        self.log.debug(
            ready_worker, "ready worker",
            len(self.ready_hosts), "ready hosts")
        hosts_in_progress = 0

        if self.config["location_group"] is not None:
            self.log.debug(
                "check if interval of the "
                "location group is exceeded")
            if self.config["location_group"]["last_check"] > time.time():
                self.config["location_group"]["last_check"] = time.time()
            elif self.config["location_group"]["last_check"] + 90 < time.time():
                self.log.info(
                    "fetching new data for the "
                    "location groups is queued")
                self.config["location_group"]["last_check"] = time.time()
                self.dispatcher.add_job({"action": "get-location-group"})

        for host_id, host in self.hosts_by_id.items():
            poll_interval = host.poll_interval

            self.log.debug(
                "check if host %s is ready, interval %s, last check %s"
                % (host.host_id, poll_interval, host.last_run))

            if host.in_progress > 0:
                self.log.info(
                    "host %s in progress since %s"
                    % (host.host_id, host.in_progress))
                hosts_in_progress += 1
                continue

            if host.in_queue:
                self.log.info(
                    "host %s waiting in queues"
                    % host.host_id)
                continue

            # Fix time issues. It's possible that the system
            # startet recently and the time was corrected via
            # NTP. If the time was corrected too often, a check
            # of the services is forced (time_fixed).
            self.log.debug("host %s last run: %s > %s" % (
                host.host_id, host.last_run, time.time()
            ))
            if host.last_run > time.time():
                self.log.warning(
                    "time fixed from %s to %s [host_id=%s]" % (
                    host.last_run, time.time(), host.host_id))
                host.last_run = time.time()
                host.time_fixed += 1
                if host.time_fixed == 15:
                    self.log.warning(
                        "too many time fixes, force host run [host_id=%s]"
                        % host.host_id)
                    host.last_run = 0
                    host.time_fixed = 0

            self.log.debug("check if host %s is ready: %s + %s (=%s) < %s" % (
                host.host_id,
                utils.timestamp(host.last_run),
                poll_interval,
                utils.timestamp(host.last_run + poll_interval),
                utils.timestamp(time.time()),
            ))
            if host.last_run + poll_interval < time.time():
                self.log.info("host %s ready" % host.host_id)
                self.ready_hosts.append(host_id)
                host.in_queue = True

        while (len(self.ready_hosts)
                and ready_worker > 0
                and hosts_in_progress < self.config["max_concurrent_hosts"]):
            host_id = self.ready_hosts.pop(0)
            host = self.hosts_by_id[host_id]
            host.in_queue = None
            host.in_progress = time.time()
            hosts_in_progress += 1
            ready_worker -= 1
            self.log.info("host %s ready for processing" % host.host_id)
            self.dispatcher.add_job({
                "action": "get-services",
                "host_id": host.host_id,
                "password": host.password,
                "proxy": host.proxy,
            })

    def on_finish(self, success, job):
        self.log.debug("finished job", success, job)
        # In case that a job failed and dependent on the
        # process state the result is adjusted.
        if success is False:
            # Just ignoring.
            if job["action"] == "get-location-group":
                return

            # the data is discarded and the host
            # is checked at the next interval
            if job["action"] == "get-services" or job["action"] == "send-data":
                self.finish_host(job["host_id"])
                return

            # set the status to UNKOWN
            if job["action"] == "service-check":
                job = {
                    "action": "service-check",
                    "service_id": job["service"]["service_id"],
                    "host_id": job["host_id"],
                    "data": {
                        "status": "UNKNOWN",
                        "message": (
                            "agent subprocess was not successful, "
                            "see logfile for more information"
                        )
                    }
                }

        if job["action"] == "get-location-group":
            return self.update_location_group(job["data"])

        # Maybe deleted by update_location_groups()
        if job["host_id"] not in self.hosts_by_id:
            return

        host = self.hosts_by_id[job["host_id"]]

        if job["action"] == "get-services":
            if not job["data"]:
                self.log.notice("no data received for host", host.host_id)
                self.finish_host(job["host_id"])
                return

            if ("poll_interval" in job["data"]
                    and host.poll_interval != job["data"]["poll_interval"]):
                if job["data"]["poll_interval"] < 60:
                    host.poll_interval = job["data"]["poll_interval"]
                else:
                    host.poll_interval = 60
                self.log.notice(
                    "adjust interval of host %s to %ss"
                    % (host.host_id, host.poll_interval))

            if not len(job["data"]["services"]):
                self.log.notice("no services ready for host", host.host_id)
                self.finish_host(job["host_id"])
                return

            host.num_services = len(job["data"]["services"])
            host.services = job["data"]["services"]
            while (len(host.services)
                    and host.services_in_progress < self.config["max_concurrent_checks"]):
                service = host.services.pop()
                host.services_in_progress += 1
                self.dispatcher.add_job({
                    "action": "service-check",
                    "service": service,
                    "host_id": job["host_id"]
                })
            return

        if job["action"] == "service-check":
            self.log.info("service data", job["data"])
            host.services_in_progress -= 1
            host.services_finished.append({
                "service_id": job["service_id"],
                "data": job["data"]
            })
            self.log.debug(
                "%s/%s services finished of host %s"
                % (len(host.services_finished), host.num_services, host.host_id))
            if len(host.services_finished) == host.num_services:
                self.dispatcher.add_job({
                    "action": "send-data",
                    "host_id": host.host_id,
                    "password": host.password,
                    "proxy": host.proxy,
                    "data": host.services_finished,
                    "location_group": host.location_group,
                })
            if len(host.services):
                service = host.services.pop()
                host.services_in_progress += 1
                self.dispatcher.add_job({
                    "action": "service-check",
                    "service": service,
                    "host_id": job["host_id"]
                })
            return

        if job["action"] == "send-data":
            self.finish_host(job["host_id"])
            return

        # Should never occur
        raise Exception("invalid job", job)

    def finish_host(self, host_id):
        host = self.hosts_by_id[host_id]
        self.log.notice("host %s finished" % host.host_id)
        host.num_services = 0
        host.services_finished.clear()
        host.services_in_progress = 0
        host.in_progress = 0
        host.last_run = time.time()

    # Executed in worker
    # Init a newly forked worker.
    def on_init(self):
        self.log.debug("init worker", os.getpid())

    # Executed in worker.
    # There exist three actions:
    #   - Request service checks from the bloonix server
    #   - Execute service checks
    #   - Send status with metrics back to the bloonix server
    def on_process(self, job):
        self.log.debug("process job", job)
        if job["action"] == "get-location-group":
            return self.get_location_group(job)
        if job["action"] == "get-services":
            return self.get_services(job)
        if job["action"] == "service-check":
            return self.run_service_check(job)
        if job["action"] == "send-data":
            return self.send_service_data(job)
        # Should never occur.
        raise Exception("invalid job", job)

    def get_location_group(self, job):
        config = self.config["location_group"]
        sock = Client(**self.config["server"])
        ret = {
            "action": "get-location-group",
            "data": None}

        try:
            sock.connect(timeout=30)
            data = {
                "action": "get-location-group",
                "location_id": config["location_id"],
                "password": config["password"],
                "version": self.VERSION}
            if config["proxy"] != None:
                data["proxy"] = config["proxy"]
            sock.send(data)
            response = sock.recv(timeout=30)
            self.log.info(response)
            Validate.server_response(response, "location_group")
            if response["status"] == "ok":
                ret["data"] = response.get("data", [])
            else:
                self.log.error(response.get("message", "request failed"))
        except ConnectionRefusedError:
            self.log.error("connection refused")
        except Exception:
            self.log.exception()

        return ret

    def get_services(self, job):
        self.log.notice(
            "request services for host %s"
            % job["host_id"])
        sock = Client(**self.config["server"])
        ret = {
            "action": "get-services",
            "data": None,
            "host_id": job["host_id"]}

        try:
            sock.connect(timeout=30, retry=2)
            data = {
                "action": "get-services",
                "host_id": job["host_id"],
                "password": job["password"],
                "version": self.VERSION}
            if job["proxy"] != None:
                data["proxy"] = job["proxy"]
            sock.send(data)
            response = sock.recv(timeout=30)
            Validate.server_response(response, "services")
            if response["status"] == "ok":
                self.log.info(
                    "services to check: %s"
                    % len(response["data"]["services"]))
                ret["data"] = response["data"]
            else:
                self.log.error(
                    "request to server failed, response:",
                    response.get("message", "unknown error"))
        except ConnectionRefusedError:
            self.log.error("connection refused")
        except Exception:
            self.log.exception()

        # We got no data or run into a exception.
        # In both cases the host is checked at the next interval.
        return ret

    def run_service_check(self, job):
        service = job["service"]
        self.log.notice(
            "check host %s service %s command %s"
            % (service["host_id"], service["service_id"], service["command"]))
        if self.test_data:
            data = self.test_data.get(service["command"], {
                "status": "OK",
                "message": "OK",
            }).copy()
            data["message"] = "%s (%s)" % (data["message"], time.time())
            #data["status"] = "CRITICAL"
        else:
            #self.log.notice({
            #    "command": service["command"],
            #    "command_options": service.get("command_options", None),
            #    "plugin_path": self.config["plugin_path"],
            #    "timeout": service.get("timeout", 60),
            #    "host_id": service["host_id"],
            #    "service_id": service["service_id"]}, dump=True)
            data = execute_check(
                command=service["command"],
                command_options=service.get("command_options", None),
                plugin_path=self.config["plugin_path"],
                plugin_cache_dir=self.config["plugin_cache_dir"],
                timeout=service.get("timeout", 60),
                host_id=service["host_id"],
                service_id=service["service_id"])
        return {
            "action": "service-check",
            "service_id": service["service_id"],
            "data": data,
            "host_id": job["host_id"]}

    def send_service_data(self, job):
        data = job["data"]
        self.log.notice("send data to server for host %s" % job["host_id"])
        self.log.debug(data)
        sock = Client(**self.config["server"])

        try:
            data = {
                "action": "service-data",
                "host_id": job["host_id"],
                "password": job["password"],
                "version": self.VERSION,
                "data": data}
            if job["proxy"] != None:
                data["proxy"] = job["proxy"]
            if job["location_group"] is not True:
                data["facts"] = get_facts()
            sock.connect(timeout=30, retry=2)
            sock.send(data)
            # The server needs some time to get
            # the data from the TCP buffer.
            try:
                data = sock.recv(timeout=15)
                self.log.debug(data)
            except Exception as err:
                self.log.warning("recv data failed: %s" % err)
        except ConnectionRefusedError:
            self.log.error("connection refused")
        except Exception:
            self.log.exception()

        # Discard data and the host is checked at the next interval.
        return {
            "action": "send-data",
            "host_id": job["host_id"]}

    def update_location_group(self, data):
        config = self.config["location_group"]
        #self.log.warning("*" * 50)
        #self.log.warning(data)
        if not data:
            self.log.notice("location group has no hosts")
            return

        self.log.notice(
            "update location group with %s hosts"
            % len(data))
        self.log.debug("data", data)
        new_hosts_by_id = {}
        for row in data:
            new_hosts_by_id[row["host_id"]] = row["password"]

        host_ids = []
        for host_id in list(self.hosts_by_id):
            host = self.hosts_by_id[host_id]
            if host.location_group is True:
                if host.host_id in new_hosts_by_id:
                    host_ids.append(host_id)
                else:
                    self.log.info("delete host", host_id)
                    del self.hosts_by_id[host_id]

        for host_id, password in new_hosts_by_id.items():
            if host_id not in host_ids:
                self.log.info("add new host", host_id)
                self.save_host({
                    "host_id": host_id,
                    "password": password,
                    "proxy": config["proxy"],
                }, True)

